pro part_c, search_tag, cal_file, out_file

; Part_C applies the calibration as derived from part_A and part_B.

; INPUT:
; search_tag = string with wildcard to identify the frames (FITS files) to be processed.
; cal_file   = file name of the part_B calibration file to be applied.

; OPTIONAL INPUT:
; out_file   = output filename [default = 'IRS2_ramp.fits'] 
; >>>>> OUTPUT FILE IS IN ADU, not electrons

;-------------------------------------------------------------------------------
st = systime(1)
file = file_search(search_tag,count=nfiles)
if (n_elements(out_file) ne 1) then out_file = file_basename(file,'.fits')+'_IRS2_ramp.fits'

restore,cal_file
;files,f5,filter_r,alpha_0,beta_0,alpha_1,beta_1,alpha_2,beta_2,alpha_3,beta_3
filter_r = 0
if ((total(beta_0) ne 0) or (total(beta_1) ne 0) or (total(beta_2) ne 0) or (total(beta_3) ne 0)) then do_beta=1 else do_beta=0
if (do_beta eq 0) then print,'Using reference pixels only'

here =  where(finite(alpha_0,/nan),nhere)
if (nhere gt 1) then alpha_0[here] = 0.
here =  where(finite(alpha_1,/nan),nhere)
if (nhere gt 1) then alpha_1[here] = 0.
here =  where(finite(alpha_2,/nan),nhere)
if (nhere gt 1) then alpha_2[here] = 0.
here =  where(finite(alpha_3,/nan),nhere)
if (nhere gt 1) then alpha_3[here] = 0.
here =  where(finite(beta_0,/nan),nhere)
if (nhere gt 1) then beta_0[here] = 0.
here =  where(finite(beta_1,/nan),nhere)
if (nhere gt 1) then beta_1[here] = 0.
here =  where(finite(beta_2,/nan),nhere)
if (nhere gt 1) then beta_2[here] = 0.
here =  where(finite(beta_3,/nan),nhere)
if (nhere gt 1) then beta_3[here] = 0.

; Read header to get data setup
head = headfits(file[0])
scipix_n = sxpar(head,'NRS_NORM')
refpix_r = sxpar(head,'NRS_REF')
print,'SCIPIX_N = ',SCIPIX_N,'   REFPIX_R = ',REFpIX_R

if (n_elements(gain)) ne 1 then begin 
  case (sxpar(head,'SCA_ID')) of 
     491  : gain = 1.0  ; new detectors May 2014 ; changes 2PM may 14
     492  : gain = 1.11
;     491  : gain = 1.24 ; old detectors
;     492  : gain = 1.21
     else : gain = 1.34
  endcase
endif
;gain = 1.24 ; electrons/DN Array 491
print,'GAIN = ',gain

; This is the effective number of pixels sampled during the pause at the end of each row. 
; The padding is needed to preserve phase of temporally periodic signals. 
; Still good for prototype_v2
;pad = 12 
; Jan 2013, prototype_v3
pad = 8
row = (scipix_n+refpix_r+2)*(512/scipix_n)+pad

print,'mark1: ',systime(1)-st,' sec     ',memory(/high)/1e9,' GB'

; read data
data0 = float(readfits(file,/silent)) ; for ramps in separated fits files
;data0 += (data0 lt 0)*2L^16
data0 *= gain

s = size(data0)
nn = lindgen(s[3])
nn -= mean(nn) ; subtract mean(nn) to minimize correlation of slope and offset

hnorm = lindgen(512)+refpix_r*((lindgen(512)+scipix_n/2) / scipix_n) ; column indices of normal pixels ; ***
href = lindgen(512/scipix_n*refpix_r)+scipix_n*(lindgen(512/scipix_n*refpix_r) / refpix_r)+scipix_n/2 ; column indices of reference pixels ; ***
hnorm1 = lindgen(512)+(refpix_r+2)*((lindgen(512)+scipix_n/2) / scipix_n) ; column indices of normal pixels ; ***
href1 = lindgen(512/scipix_n*refpix_r)+(scipix_n+2)*(lindgen(512/scipix_n*refpix_r) / refpix_r)+(scipix_n/2+1) ; column indices of reference pixels ; ***

; remove linear trends per pixel (gains, offsets)
data0 = reform(data0,s[1]*s[2],s[3],/over)
sxy = [nn##data0[0:1999,*],nn##data0[2000:*,*]] ; nn##data ; For obscure IDL reasons, this caluclation is faster when done in two parts.
sx = total(nn)
sxx = total(nn^2)
sy = [(nn*0+1)##data0[0:1999,*],(nn*0+1)##data0[2000:*,*]] ; total(data,2)
;m = (s[3]*sxy-sx*sy)/(s[3]*sxx-sx^2) ; ***
b = (sy*sxx-sxy*sx)/(s[3]*sxx-sx^2) ; ***
;  for i=0,s[3]-1 do data0[*,i] -= nn[i]*m+b  ; subtract slope and offset
for i=0,s[3]-1 do data0[*,i] -= b ; subtract the offset only
sxy = 0
sy = 0
b = 0
data0 = reform(data0,s[1],s[2],s[3],/over)
print,'mark2: ',systime(1)-st,' sec     ',memory(/high)/1e9,' GB'

; reform the data
data0 = reform(data0,s[1]/5,5,s[2],s[3],/over)
data0 = transpose(data0,[0,2,3,1])
rev = reverse(lindgen(s[1]/5))
data0[*,*,*,0] = data0[rev,*,*,0]
data0[*,*,*,2] = data0[rev,*,*,2]
data0[*,*,*,4] = data0[rev,*,*,4]

; convert to time sequences of normal pixels and reference pixels.
d0 = fltarr(s[1]/5+pad+2*(512/scipix_n),s[2],s[3],5)
d0[hnorm1,*,*,*] = data0[hnorm,*,*,*]
d0[href1,*,*,*] = data0[href,*,*,*]
data0 = temporary(d0)

;<<<<< Fitting and removal of slopes per frame to remove issues at frame boundaries.
time = findgen(row,s[2])
row4plus4 = [0,1,2,3,2044,2045,2046,2047]
time = time-mean(time)
ab_3 = fltarr(2,s[3],5)  ; for top+bottom ref pixel rows
for i=0,4 do begin 
  for k=0,s[3]-1 do begin
    ab_3[*,k,i] = linfit((time[*,row4plus4])[where(data0[*,row4plus4,k,i] ne 0)],(data0[*,row4plus4,k,i])[where(data0[*,row4plus4,k,i] ne 0)])
  endfor
endfor

for i=0,4 do begin 
  for k=0,s[3]-1 do data0[*,*,k,i] -= (time*ab_3[1,k,i]+ab_3[0,k,i])*(data0[*,*,k,i] ne 0)
endfor
time = 0
;<<<<<<<

print,'mark3: ',systime(1)-st,' sec     ',memory(/high)/1e9,' GB'
; Use cosine weighted interpolation to replace 0.0 values at bad pixels and gaps. (initial guess)
  w = (sin((findgen(32)/32)*!pi))[1:*]
;  for kk=0,3 do begin
  for kk=0,0 do begin
    for jj=0L,s[3]-1 do begin
      dat = reform(data0[*,*,jj,kk],row*s[2])
      mask = float(dat ne 0)
      dat = reform(convol(dat,w,/edge_wrap)/convol(mask,w,/edge_wrap),row,s[2])
      data0[*,*,jj,kk] += dat*(1-mask)
    endfor
  endfor
  here = where(finite(data0,/nan),nhere)
  if (nhere ge 1) then data0[here] = 0
  print,'mark3a: ',systime(1)-st,' sec     ',memory(/high)/1e9,' GB'

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Use Fourier filter/interpolation to replace  
; (a) bad pixel, gaps, and reference data in the time-ordered normal data
; (b) gaps and normal data in the time-ordered reference data
; This "improves" upon the cosine interpolation performed above.

; This process might also be made more memory efficient by using shared memory
; (SHMMAP, SHMVAR) instead of passing data0, but at this point in the code 
; avoiding a temprary duplication of the data is not critical. Shared memory
; might also improve speed may avoiding the copying, but this is probably only
; a matter of a few seconds.

; parameters for the filter to be used
  elen=110000L/(scipix_n+refpix_r+2)    ; length of apodization cosine filter
  blen=(512L+512L/scipix_n*(refpix_r+2)+pad)/(scipix_n+refpix_r+2)*s[2]/2-elen/2 ; max unfiltered frequency 

; construct the filter [1,cos,0,cos,1]
  a=(cos(findgen(elen)*!pi/elen)+1)/2 
  a=[replicate(1.,blen),a,fltarr(row*s[2]/2-2*blen-2*elen),reverse(a),replicate(1.,blen)]
  a=[a,reverse(shift(a,-1))]
  aa = a#replicate(1,s[3]) ; for application to the data
  afilter = a

  n_iter_norm = 3
  dd0 = data0[*,*,*,0]
fft_interp_norm,dd0,2,replicate(1,s[1]/4,s[2],4),row,hnorm,hnorm1,s,aa,n_iter_norm
data0[*,*,*,0] = dd0
  print,'mark3b: ',systime(1)-st,' sec     ',memory(/high)/1e9,' GB'
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; indices for keeping/shuffling reference pixels
ht = indgen(scipix_n+refpix_r+2,512/scipix_n)
ht[0:(scipix_n-refpix_r)/2,*]=-1
ht[scipix_n/2+1+3*refpix_r/2:*,*]=-1 
hs=ht 
ht=reform(ht[where(ht ge 0)],2*refpix_r*512/scipix_n) ; t = like href1, but extended over gaps and 1st and last norm pix.
hs[scipix_n/2+1-refpix_r/2:scipix_n/2+refpix_r+refpix_r/2,*]=hs[reform([transpose(reform(indgen(refpix_r),refpix_r/2,2)),transpose(reform(indgen(refpix_r),refpix_r/2,2))],refpix_r*2)+scipix_n/2+1,*]  ; WIRED for R=2^(int)
hs=reform(hs[where(hs ge 0)],2*refpix_r*512/scipix_n) ; s = like href1, shuffled and replicated
if (refpix_r mod 4 eq 2) then hs = reform(reverse(reform(hs,2,n_elements(hs)/2)),n_elements(hs))

; construct the reference data
r0 = data0
r0[*] = 0
r0[ht,*,*,*] = data0[hs,*,*,*]
print,'mark4: ',systime(1)-st,' sec     ',memory(/high)/1e9,' GB'
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

sd = size(data0)
if (do_beta) then begin  
  refout0 = reform(data0[*,*,*,0],sd[1]*sd[2],sd[3])
  refout0 = fft(refout0,dim=1,/over)
endif
r0 = reform(r0,sd[1]*sd[2],sd[3],5,/over)
r0 = complex(r0)
index = [1,2,3,4]
oBridge = objarr(4)
for k=0,3 do begin
  oBridge[k] = OBJ_NEW("IDL_IDLBridge")
  oBridge[k]->SetVar, "r0", r0[*,*,index[k]]
  oBridge[k]->Execute, "r0 = fft(r0,dim=1,/overwrite)", /nowait
endfor
repeat begin
  status = lonarr(4)
  for i=0,3 do status[i] = oBridge[i]->Status()
  print,systime(1)-st,status,string(13b),form='(f,4i8,a,$)'
  spawn,'sleep 1'
endrep until (total(status) eq 0)
print
print,'mark5: ',systime(1)-st,' sec     ',memory(/high)/1e9,' GB'

oBridge[0]->SetVar, "alpha", alpha_0
oBridge[1]->SetVar, "alpha", alpha_1
oBridge[2]->SetVar, "alpha", alpha_2
oBridge[3]->SetVar, "alpha", alpha_3
for k=0,3 do oBridge[k]->SetVar, "s3", s[3]
for k=0,3 do oBridge[k]->Execute, "for i=0,s3-1 do r0[*,i] *= alpha"
for k=0,3 do oBridge[k]->SetVar, "alpha", 0
if (do_beta) then begin
  oBridge[0]->SetVar, "beta", beta_0
  oBridge[1]->SetVar, "beta", beta_1
  oBridge[2]->SetVar, "beta", beta_2
  oBridge[3]->SetVar, "beta", beta_3
  for k=0,3 do oBridge[k]->SetVar, "refout0", refout0
  for k=0,3 do oBridge[k]->Execute, "for i=0,s3-1 do r0[*,i] += beta*refout0[*,i]"
  for k=0,3 do oBridge[k]->SetVar, "refout0", 0
  for k=0,3 do oBridge[k]->SetVar, "beta", 0
endif

for k=0,3 do oBridge[k]->Execute, "r0 = fft(r0,1,dim=1,/overwrite)", /nowait
repeat begin
  status = lonarr(4)
  for i=0,3 do status[i] = oBridge[i]->Status()
  print,systime(1)-st,status,string(13b),form='(f,4i8,a,$)'
  spawn,'sleep 1'
endrep until (total(status) eq 0)
print
r0[*,*,1] = oBridge[0]->GetVar('r0')
r0[*,*,2] = oBridge[1]->GetVar('r0')
r0[*,*,3] = oBridge[2]->GetVar('r0')
r0[*,*,4] = oBridge[3]->GetVar('r0')
for k=0,3 do obj_destroy,oBridge[k]

r0 = reform(r0,sd[1],sd[2],sd[3],5,/over)
r0 = float(r0)
r0 = r0[hnorm1,*,*,*]

print,'mark6: ',systime(1)-st,' sec     ',memory(/high)/1e9,' GB'

rev = reverse(lindgen(512))
data0 = data0[hnorm1,*,*,*]
data0 -= r0
data0[*,*,*,2] = data0[rev,*,*,2]
data0[*,*,*,4] = data0[rev,*,*,4]
data0 = transpose(data0,[0,3,1,2])
data0 = reform(data0[*,1:*,*,*],s[2],s[2],s[3],/over)
data0 /= gain
sxaddpar,head,'BZERO',0.0
writefits,out_file,data0,head
print,'mark7: ',systime(1)-st,' sec     ',memory(/high)/1e9,' GB'

end
